#include <error.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <pthread.h>

#include "dormirUnPeu.h"

// Nombre de trains dans le circuit
#define NB_TRAINS 2

// Nombre de tours de montagne russe que fait chaque train
#define NB_TOURS 5

pthread_mutex_t zoneE = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t zoneH = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t zoneA = PTHREAD_MUTEX_INITIALIZER;

void *codeTrain(void *arg){
  int nbTours, rc;
  int numTrain = *(int *)arg;
  free(arg);

  printf("Train %d : Je suis dans la zone zoneG\n", numTrain);
  dormirUnPeu();
  
  // P(zoneE);
  rc = pthread_mutex_lock(&zoneE);
  if (rc != 0)
    error_at_line(EXIT_FAILURE, rc, __FILE__, __LINE__, "pthread_mutex_lock");

  for (nbTours=0 ; nbTours<NB_TOURS ; nbTours++){
    printf("Train %d : Je suis dans la zone zoneE\n", numTrain);
    dormirUnPeu();
    // P(zoneH);
    rc = pthread_mutex_lock(&zoneH);
    if (rc != 0)
      error_at_line(EXIT_FAILURE, rc, __FILE__, __LINE__, "pthread_mutex_lock");
    // V(zoneE);
    rc = pthread_mutex_unlock(&zoneE);
    if (rc != 0)
      error_at_line(EXIT_FAILURE, rc, __FILE__, __LINE__, "pthread_mutex_unlock");
    
    printf("Train %d : Je suis dans la zone zoneH\n", numTrain);
    dormirUnPeu();
    // P(zoneA);
    rc = pthread_mutex_lock(&zoneA);
    if (rc != 0)
      error_at_line(EXIT_FAILURE, rc, __FILE__, __LINE__, "pthread_mutex_lock");
    // V(zoneH);
    rc = pthread_mutex_unlock(&zoneH);
    if (rc != 0)
      error_at_line(EXIT_FAILURE, rc, __FILE__, __LINE__, "pthread_mutex_unlock");

    printf("Train %d : Je suis dans la zone zoneA\n", numTrain);
    dormirUnPeu();
    // P(zoneE);
    rc = pthread_mutex_lock(&zoneE);
    if (rc != 0)
      error_at_line(EXIT_FAILURE, rc, __FILE__, __LINE__, "pthread_mutex_lock");
    // V(zoneA);
    rc = pthread_mutex_unlock(&zoneA);
    if (rc != 0)
      error_at_line(EXIT_FAILURE, rc, __FILE__, __LINE__, "pthread_mutex_unlock");
  }

  // V(zoneE);
  rc = pthread_mutex_unlock(&zoneE);
  if (rc != 0)
    error_at_line(EXIT_FAILURE, rc, __FILE__, __LINE__, "pthread_mutex_unlock");
  printf("Train %d : J'ai fini de tourner\n", numTrain);

  pthread_exit((void*)0);
}

int main() {
  int numThread, rc;

  for (numThread=0; numThread<NB_TRAINS; numThread++){
    pthread_t thread;
    int *pNum = malloc(sizeof(int));
    if (pNum == NULL)
      error_at_line(EXIT_FAILURE, 0, __FILE__, __LINE__,
		    "malloc a des problemes (probablement plus de memoire)");
    *pNum = numThread;
    rc = pthread_create(&thread, NULL, codeTrain, (void *)pNum);
    if (rc != 0)
      error_at_line(EXIT_FAILURE, rc, __FILE__, __LINE__, "pthread_create");
  }

  //On attend que tous les threads aient fini de s'excuter
  sleep(20);

  // On detruit tous les objets de synchronisation
  rc = pthread_mutex_destroy(&zoneE);
  if (rc != 0)
    error_at_line(EXIT_FAILURE, rc, __FILE__, __LINE__, "pthread_mutex_destroy");
  rc = pthread_mutex_destroy(&zoneH);
  if (rc != 0)
    error_at_line(EXIT_FAILURE, rc, __FILE__, __LINE__, "pthread_mutex_destroy");
  rc = pthread_mutex_destroy(&zoneA);
  if (rc != 0)
    error_at_line(EXIT_FAILURE, rc, __FILE__, __LINE__, "pthread_mutex_destroy");

  printf("Main() termine\n");

  exit(EXIT_SUCCESS);
}
